implementation module Prelude

from StdOverloaded import 
     class ==(..),  class <(..),    class +(..),    class -(..),    class *(..),    class /(..)
    ,class abs(..), class zero(..), class one(..),  class exp(..),  class ln(..)
    ,class sin(..), class cos(..),  class asin(..), class acos(..), class atan(..)
    ,class sinh(..),class cosh(..), class asinh(..),class acosh(..), class atanh(..)
    ,class fromInt(..),             class sign(..), class toChar(..),               class fromChar(..)
    ,class toReal(..), class fromReal(..)

from StdBool        import  not,||,&&,instance == Bool
from StdFunc        import  o
from StdMisc        import  abort

import  intInstances
import realInstances 
import charInstances

from Ratio          import  :: Rational, :: Ratio(..)
from PreludeLayer   import  class hs__quot(..), class hs__rem(..),  instance hs__quot Int, instance hs__rem Int
						  , from2,from_to,from_then_to,toStr
from PreludeText    import  class Show(..), :: ShowS(..)
from PreludeList    import  foldr,repeat,concat,iterate,takeWhile,map


// =====================================================================================================

// -------- EQ class ----------- 

class Eq a | == a
where
    (/=) infix  4 :: !a !a  ->  Bool | Eq a
    (/=) x y :== not (x == y)

// -------- ORD class ----------- 

class Ord a | Eq, < a
where
    compare :: !a !a -> Ordering   | Ord a
    compare x y :== if (x==y) EQ (if (x<y) LT GT)
                                
    (>) infix  4 :: !a !a   ->  Bool | Ord a
    (>) x y  :== y < x 
    
    (<=) infix  4 :: !a !a  ->  Bool | Ord a
    (<=) x y :== not (y<x)
    
    (>=) infix  4 :: !a !a  ->  Bool | Ord a
    (>=) x y :== not (x<y) 
    
    min::!a !a ->   a | Ord a
    min x y  :== if (x<y) x y
    
    max::!a !a ->   a | Ord a
    max x y  :== if (x<y) y x

// -------- ENUM class -----------    

class Enum a   
where
    toEnum          :: Int -> a
    fromEnum        :: a -> Int
    enumFromThen    :: a a -> [a]                       // [n,n`..]
    
    succ    :: !a -> a | Enum a
    succ  x :== toEnum ((fromEnum x) + 1)
    
    pred    :: !a -> a | Enum a
    pred  x :== toEnum ((fromEnum x) - 1)

    enumFrom :: !a -> [a]  | Enum a                     // [n..]
    enumFrom x :== map toEnum (from2 (fromEnum x)) 
    
    enumFromTo :: !a !a -> [a] | Enum a                 // [n..m]
    enumFromTo x y :== map toEnum (from_to (fromEnum x) (fromEnum y))
    
    enumFromThenTo :: !a !a !a -> [a] | Enum a          // [n,n`..m]
    enumFromThenTo x y z :== map toEnum (from_then_to (fromEnum x) (fromEnum y) (fromEnum z))

// -------- BOUNDED class ----------- 

class Bounded a 
where
    minBound :: a
    maxBound :: a

// -------- NUM class ----------- 

class Num a | Eq,+,-,*,abs,fromInt,zero,one,Show a
where
    signum      :: a -> a

    negate      ::  a -> a  | Num a
    negate x    :== zero - x

/*  fromInteger     :: Integer -> a
                not supported yet */
// -------- REAL class -----------
class Hs__Real a | Num, Ord a                   //shoulc be Real : name-space bug clean 2.0
where
    toRational :: a -> Rational

// -------- INTEGRAL class -----------

class  Integral a | Hs__Real,Enum a
where
    quotRem     :: a a -> (a,a)
    toInteger   :: a -> Int                     
    
    (quot)  infixl 7:: a a -> a | Integral a    // `quot` is integer division truncated toward zero
    (quot) n d      :==  fst (quotRem n d)
    
    (rem) infixl 7  :: a a -> a | Integral a
    (rem) n d       :==  snd (quotRem n d)
    
    (div) infixl 7  :: a a -> a | Integral a    // `div` is integer division truncated toward negative infinity
    (div) n d       :==  fst (divMod n d)
    
    (mod) infixl 7  :: a a -> a | Integral a
    (mod) n d       :==  snd (divMod n d)
    
    divMod          :: a a -> (a,a) | Integral a
    divMod n d      :==  
        let (q,r) = quotRem n d
        in
            if ( signum r == ( negate (signum d) )) (q-one, r+d) (q,r)

// -------- FRACTIONAL class -----------    

class Fractional a  | Num,/ a
where
    fromRational     :: Rational -> a
    recip   :: a -> a   | Fractional a
    recip x :==  one / x
   
// -------- FLOATING class -----------    

half :: a | one,/,+ a 
half        = one / (one+one)

minusone :: a | zero,-,one a
minusone    = zero - one

two :: a | one, + a
two         = one + one

class Floating a | Fractional,exp,sin,cos,asin,acos,atan,sinh,cosh,asinh,acosh,atanh a
where
    pi              :: a
    log             :: a -> a 
    
    (**) infixr 8   :: a a -> a     | Floating a
    (**) x y        :== exp (log x * y)
    
    tan             :: a -> a       | Floating a
    tan  x          :== sin x / cos x
    
    logBase         :: a a -> a     | Floating a
    logBase x y     :== log y / log x
    
    sqrt            :: !a -> a      | Floating a 
    sqrt x          :==  x ** half
        
    tanh            :: a -> a       | Floating a
    tanh x          :==  sinh x / cosh x
  
// -------- REALFRAC class -----------    

class RealFrac a | Hs__Real,Fractional a
where
    properFraction :: a -> (b,a)    | Integral b
    
    truncate    :: a -> b       | RealFrac a & Integral b
    truncate x  :== fst (properFraction x)
    
    round  :: a -> b        | RealFrac a & Integral b
    round x :==
        let
            (n,r) = properFraction x
            m = if (r < zero) (n - one) (n + one)
        in 
            case signum (abs r - half) of
                minusone        -> n
                zero            -> if (even n) n m
                one             -> m

    ceiling :: a -> b       | RealFrac a & Integral b
    ceiling x :== let (n,r) = properFraction x in
                        if (r > zero) (n + one) n

    floor   :: a -> b | RealFrac a & Integral b
    floor x :== let (n,r) = properFraction x in 
                if (r < zero) (n - one) n
      
// -------- REALFLOAT class -----------    
    
class RealFloat a | RealFrac,Floating a
where
    floatRadix      :: a -> Int
    floatDigits     :: a -> Int
    floatRange      :: a -> (Int,Int)
    decodeFloat     :: a -> (Int,Int)
    encodeFloat     :: Int Int -> a
    
    isNaN           :: a -> Bool
    isInfinite      :: a -> Bool
    isDenormalized  :: a -> Bool
    isNegativeZero  :: a -> Bool
    isIEEE          :: a -> Bool
    
    exponent    :: a -> Int | RealFloat a
    exponent x :==
        let (m,n) = decodeFloat x in
        if (m == 0) 0 (n + floatDigits x)
                        
    significand :: a -> a       | RealFloat a
    significand x :==
        let (m,_) = decodeFloat x in
        encodeFloat m (negate floatDigits x)
                        
    scaleFloat :: Int a -> a    | RealFloat a
    scaleFloat k x :==
        let (m,n) = decodeFloat x in
        encodeFloat m (n+k)
                        
    atan2       :: a a -> a | RealFloat a
    atan2 y x   :== atan2` y x

atan2` :: a a -> a | RealFloat a 
atan2` y x   
    | (x > zero)                = atan (y/x)
    | x == zero && y > zero     =  pi / two
    | x <  zero && y > zero     =  pi + atan (y/x) 
    |       (x <= zero && y < zero)
        ||  (x <  zero && isNegativeZero y)
        ||  (isNegativeZero x && isNegativeZero y)
                                = negate (atan2 (negate y) x)
    | y == zero && ( x < zero || isNegativeZero x)
                                =  pi    // must be after the previous test on zero y
    | x == zero  && y== zero    =  y     // must be after the other double zero tests
                                =  x + y // x or y is a NaN, return a NaN (via +)

// ----------- Numeric functions ---------------

subtract :: a a -> a | Num a 
subtract a b =  flip (-) a b

even :: !a -> Bool | Integral a 
even n = n rem two == zero 

odd :: !a -> Bool | Integral a 
odd n = not (even n)

gcd :: !a !a -> a   | Integral a
gcd x y | x == y && x == zero   =  abort "gcd 0 0 is undefined"
gcd x y =  gcd` (abs x) (abs y)
    where 
        gcd` v z |  z == zero   =  v
                                =  gcd` z (v rem z)

lcm :: !a !a -> a   | Integral a
lcm z _     | z == zero =  zero
lcm _ z     | z == zero =  zero
lcm x y     =  abs ((x quot (gcd x y)) * y)


(^) infixr 8 :: !a !b -> a  | Num a & Integral b
(^) _ n | n == zero = one
        | n <  zero = abort "^: negative exponent"
(^) x n | n >  zero =  f x (n-one) x
    where
        f _ z y | z == zero = y
        f x n y             = g x n  
        where
            g x n | even n  = g (x*x) (n quot two)
                            = f x (n-one) (x*y)

(^^) infixr 8 :: !a !b -> a | Fractional a & Integral b
(^^) x n    = if (n >= zero) (x^n) (recip (x^(negate n)))

fromIntegral ::  !a -> b    | Integral a & Num b
fromIntegral n = fromInt (toInteger n) 

realToFrac  :: !a -> b  | Hs__Real a & Fractional b
realToFrac n = fromRational (toRational n)

// --------------- MONADIC classes ---------------------

class  Functor f
where
    fmap :: (a -> b) (f a) -> f b

class  Monad m
where
    (>>=) infixl 1  :: (m a) (a -> m b)     -> m b
    return :: a -> m a

    fail :: [Char] -> m a

    (>>) infixl 1 :: (m a) (m b)-> m b  | Monad m
    (>>) m k  :==  m >>= (\_ -> k)

sequence :: [m a] -> m [a]  | Monad m 
sequence mxs =  foldr mcons (return []) mxs
    where mcons p q = p >>= \x -> q >>= \y -> return [x:y]

sequence_ :: [m a] -> m Trivial | Monad m 
sequence_ mxs = foldr (>>) (return Trivial) mxs

// The xxxM functions take list arguments, but lift the function or list element to a monad type

mapM :: (a -> m b) [a] -> m [b] | Monad m
mapM f as =  sequence (map f as)

mapM_ :: (a -> m b) [a] -> m Trivial | Monad m
mapM_ f as =  sequence_ (map f as)

(=<<) infixr 1 :: (a -> m b) (m a) -> m b | Monad m
(=<<) f x =  x >>= f

id :: !a -> a
id x =  x

const :: !a b -> a
const x _  =  x

/* 
"." is not a legal operator name in Clean, automatically parsed as "o"
(.) infixr 9 :: (b -> c) (a -> b) -> (a -> c)
(.) f g =  \ x -> f (g x)
*/

flip :: (a b -> c) b a -> c
flip f x y  =  f y x

(seq) infixr 0 :: !a  b -> b
(seq) x y = y 

($) infixr 0 :: (a -> b) a -> b
($) f x =  f x

($!) infixr 0 :: (a -> b) !a -> b
($!) f x = x seq f x

// ----- Trivial data type ---------- 

:: Trivial = Trivial 

instance == Trivial where (==) x y = True
instance <  Trivial where (<)  x y = False

instance Enum Trivial
where
    toEnum  _ = Trivial
    fromEnum _ = 0
    enumFromThen _ _ = repeat Trivial

instance Bounded Trivial where
    maxBound    = Trivial
    minBound    = Trivial

// --------- BOOL type ------------

//:: Bool  =  False | True

fromBool :: Bool -> Int
fromBool False  =   0
fromBool True   =   1

toBool :: Int -> Bool
toBool 0    =   False
toBool _    =   True

instance < Bool     where   (<) x y  =  (fromBool x) <  (fromBool y)

instance Enum Bool
where
    toEnum  o   = toBool o
    fromEnum o =    fromBool o
    enumFromThen o o` 
        = map toEnum ( from_then_to (fromEnum o) (fromEnum o`) (fromEnum lastElem))
        where 
            lastElem :: Bool
            lastElem | o` < o   = minBound
                                = maxBound

instance Bounded Bool where
    minBound = False
    maxBound = True

// -------------- CHAR type -----------------------

instance Enum Char
where
    toEnum c    = toChar c
    fromEnum c  = fromChar c
//  in Haskell: enumFrom c        = map toEnum [fromEnum c .. fromEnum (maxBound::Char)]
    enumFromThen c c` 
        = map toEnum ( from_then_to (fromEnum c) (fromEnum c`) (fromEnum lastElem))
        where 
            lastElem :: Char
            lastElem | c` < c   = minBound
                                = maxBound

instance  Bounded Char  where
    minBound    =  '\0'
    maxBound    =  '\255'


// -------------- MAYBE type -----------------------

:: Maybe a  =   Just a
            |   Nothing

instance == (Maybe a) | == a
where 
    (==) Nothing Nothing    = True
    (==) (Just x) (Just y)  = x == y
    (==) _       _          = False

instance < (Maybe a) | < a
where 
    (<) Nothing (Just x)    = True
    (<) (Just x) (Just y)   = x < y
    (<) _        _          = False

/* See in PreludeGeneric:
instance Show  (Maybe a) | Show a 
instance Read  (Maybe a) | Read a 
*/

maybe :: b (a -> b) (Maybe a) -> b
maybe n f Nothing  =  n
maybe n f (Just x) =  f x

instance  Functor Maybe
where
    fmap f Nothing    =  Nothing
    fmap f (Just x)   =  Just (f x)
        

instance Monad Maybe
where
    (>>=) Nothing  k = Nothing
    (>>=) (Just x) k = k x 
    
    return x    =  Just x
    fail s      =  Nothing

// -------------- EITHER type -----------------------

:: Either a b   =  Left a 
                | Right b
/* See in PreludeGeneric:
instance ==    (Either a b)  | == a & == b
instance <     (Either a b)  | < a & < b
instance Show  (Either a b)  | Show a & Show b
instance Read  (Either a b)  | Read a & Read b
*/

either :: (a -> c) (b -> c) (Either a b) -> c
either f g (Left x)  =  f x
either f g (Right y) =  g y

/*
// -------------- IO type -----------------------
... can be found in PreludeIO

*/

// ------------- ORDERING type ---------------------
:: Ordering = LT | EQ | GT

fromOrdering :: Ordering -> Int
fromOrdering LT =   0
fromOrdering EQ =   1
fromOrdering GT =   2

toOrdering :: Int -> Ordering
toOrdering 0    =   LT
toOrdering 1    =   EQ
toOrdering _    =   GT

instance == Ordering
where   (==) x y =  (fromOrdering x) == (fromOrdering y)

instance < Ordering
where   (<) x y  =  (fromOrdering x) <  (fromOrdering y)

instance Enum Ordering
where
    toEnum  o   = toOrdering o
    fromEnum o =    fromOrdering o
    enumFromThen o o` 
        = map toEnum ( from_then_to (fromEnum o) (fromEnum o`) (fromEnum lastElem))
        where 
            lastElem :: Ordering
            lastElem | o` < o   = minBound
                                = maxBound
instance Bounded Ordering where
    minBound = LT
    maxBound = GT

// ============================== INSTANCES FOR BASIC CLASSES =============================================

// --- Int ---
// Eq, Ord

instance Num Int
where
    signum x = sign x

instance Hs__Real Int
where 
    toRational :: Int -> Rational
    toRational n = (:%) n 1

instance Enum Int
where
    toEnum n    =   n
    fromEnum n  =   n
    enumFromThen n m = [n : enumFromThen m (m+(m-n)) ]

instance Integral Int
where
    quotRem x y = (x hs__quot y, x hs__rem y)
    toInteger n = n

instance Bounded Int    // 32 bits
    where
        minBound = (-2147483647)
        maxBound =   2147483648

// --- Real (Float) ---
// Eq, Ord

instance Num Real
where
    signum n = toReal (sign n)

instance Hs__Real Real
where
    toRational :: Real -> Rational
    toRational r = toRat r 1
        where
            toRat :: Real Int -> Rational
            toRat r denomin
                | (r - toReal ri) == 0.0    =   (ri / gcdrd) :% (denomin / gcdrd )
                                            =   toRat (r*10.0) (denomin*10)
                where
                    ri      =   toInt r
                    gcdrd   =   gcd ri denomin

            toInt :: Real -> Int
            toInt r = fromReal r

instance Fractional Real
where
    fromRational :: Rational -> Real
    fromRational (n :% m) = toReal (n/m)

instance Floating Real
where
    pi  =   3.1415926535897932384626433832795
    log x = ln x

instance RealFrac Real
where
    properFraction :: Real -> (b,Real) | Integral b
    properFraction r
        | abs r < abs (toReal fromRealr)
            |   r > 0.0
                =   (toEnum (fromRealr-1),  r - toReal (fromRealr-1))
                =   (toEnum (fromRealr+1),  r - toReal (fromRealr+1))

        =   (toEnum fromRealr,      r - toReal fromRealr )
        where
            fromRealr   :: Int
            fromRealr   =   fromReal r

// !!! instance RealFloat Real

instance  Enum Real
where
    toEnum f            = fromIntegral f
    fromEnum f          = (fromInt o truncate) f    // may overflow
    enumFromThen f1 f2  = numericEnumFromThen f1 f2

numericEnumFrom :: a -> [a] | Fractional a
numericEnumFrom x = iterate ((+) one) x

numericEnumFromThen :: a a -> [a] | Fractional a
numericEnumFromThen n m = iterate ((+) (m-n)) n

numericEnumFromTo :: a a -> [a] | Fractional, Ord a
numericEnumFromTo n m = takeWhile ((>=) (m + half) ) (numericEnumFrom n)

numericEnumFromThenTo :: a a a -> [a] | Fractional, Ord a
numericEnumFromThenTo n n` m 
    = takeWhile p (numericEnumFromThen n n`)
    where
        p   | n` > n    = ((>=) (m + (n`-n)/two))
            | otherwise = ((<=) (m + (n`-n)/two))

numericSucc :: a -> a | +, one a
numericSucc x = x + one

numericPred :: a -> a | -, one a
numericPred x = x - one

// -------- LIST -----------------------

instance == [a] | Eq a
    where
    (==) :: ![a]    ![a]    ->  Bool | Eq a
    (==) [] []          = True
    (==) [] _           = False
    (==) [_:_] []       = False
    (==) [a:as] [b:bs]
        | a == b        = as == bs
                        = False

instance < [a] | Ord a
    where
    (<) :: ![a] ![a] -> Bool | Ord a
    (<) []   []         = False
    (<) [] _            = True
    (<) [_:_] []        = False
    (<) [a:as] [b:bs]   
        | a < b         = True
        | a > b         = False
                        = as < bs

instance Functor [] where
    fmap f xs = map f xs

instance  Monad []  where
    (>>=) m k       = concat (map k m)
    return x        = [x]
    fail s          = []

// ----------------------------------------

/* See in PreludeGeneric:
instance Eq  (a,b)
instance Ord (a,b)
instance Eq  (a,b,c)
instance Ord (a,b,c)
*/

instance Bounded (a,b) | Bounded a & Bounded b
where
	minBound = (minBound,minBound)
	maxBound = (maxBound,maxBound)
	
instance Bounded (a,b,c) | Bounded a & Bounded b & Bounded c
where
	minBound = (minBound,minBound,minBound)
	maxBound = (maxBound,maxBound,maxBound)
	
// --------------------------------------------------------------------------------------

// fst      :: !(!.a,.b) -> .a
fst tuple :== t1 where (t1, _) = tuple
// snd      :: !(.a,!.b) -> .b
snd tuple :== t2 where (_, t2) = tuple

curry   :: ((a, b) -> c) a b -> c           // the same as in Clean
curry f x y      =  f (x, y)

uncurry :: (a b -> c) (a, b) -> c
uncurry f (x,y)      =  f x y


// until p f  yields the result of applying f until p holds.
until :: !(a -> Bool) (a -> a) a -> a
until p f x 
    | p x   =  x
            =  until p f (f x)

// asTypeOf is a type-restricted version of const.  It is usually used
// as an infix operator, and its typing forces its first argument
// (which is usually overloaded) to have the same type as the second.

asTypeOf  :: !a a -> a
asTypeOf x y =  const x y

//error stops execution and displays an error message

error :: [Char] -> a
error msg   =  abort (toStr msg)

// It is expected that compilers will recognize this and insert error
// messages that are more appropriate to the context in which undefined 
// appears. 

undefined   :: a
undefined   =  error ['Prelude.undefined']

// ====================== END OF PRELUDE ==================================================
