implementation module PreludeGeneric

from StdBool		import not, &&, ||
from StdList		import ++, instance fromString [x], instance toString [x]

from Prelude 		import :: Either(..), :: Ordering(..), :: Maybe(..), class fromString(..), class toString(..)
							, instance == [a]
from PreludeText 	import class Show(..), class Read(..), ::ReadS, ::ShowS, show, showString, show_
						   ,read, reads, lex, showList, showList_, readList, readList_
from Char			import lexLitChar, readLitChar

import  intInstances
import charInstances
// ------------ Types used to encode the generic representation type of kind * ---------------------

:: Unit 			= Unit						// unit type; empty
:: Con a 			= Con String a				// Constructor
:: Pair a b 		= Pair a b					// product
//:: Either a b 		= Left a | Right b			// disjoint sum

class toGen source gen :: source -> gen

// ------ Instances of Eq --------

instance == Unit
where
	(==)  _ _ = True

instance == (Con a)    | Eq a
where
	(==) (Con _ x) (Con _ y) 		=  x == y		// no test on equality of strings

instance == (Pair a b) | Eq a & Eq b
where
	(==) (Pair x1 y1) (Pair x2 y2)	= (x1 == x2) && (y1 == y2)

instance == (Either a b) | Eq a & Eq b
where 
	(==) (Left x)	 (Left y)	= x == y 	
	(==) (Right x)   (Right y)	= x == y
	(==) _			 _			= False

// ------ Instances of Ord --------
instance < Unit
where
	(<)  _ _ = True

instance < (Con a) | Ord a
where
	(<) (Con _ x) (Con _ y) 		=  x < y

instance < (Pair a b) | Ord a & Ord b 
where
	(<) (Pair x1 y1) (Pair x2 y2) = f (x1,y1) (x2,y2)
	where
		f (x1,y1) (x2,y2)
		  |  x1 == x2	= y1 < y2
						= x1 < x2

instance < (Either a b) | Ord a & Ord b
where 
	(<) (Left x)	 (Left y)	= x < y
	(<) (Right x)    (Right y)	= x < y
	(<) (Left _)	 (Right _)	= True
	(<) (Right _)	 (Left _)	= False

// ------ Instances of Show --------
instance Show Unit
where
	showsPrec _ Unit = \cs -> cs

instance Show (Con a)    | Show a
where
	showsPrec n (Con name x) = showString( openbr ++ fromString name ++ [' '] ++ show x ++ closebr)
		where
			openbr  | n > 0 	= ['(']
								= []
			closebr | n > 0 	= [')']
								= []

instance Show (Pair a b) | Show a & Show b
where
	showsPrec _ (Pair x y) = showString( show x ++ [' '] ++ show y)

instance Show (Either a b) | Show a & Show b
where
	showsPrec _ (Left  a) = showString (show a)
	showsPrec _ (Right b) = showString (show b)

// ------ Instances of Read --------
instance Read Unit
where
	readsPrec _ = \cs -> [(Unit,cs)]

instance Read (Con a) | Read a
where
	readsPrec 0 = \cs -> [(Con (toString name) x,tail_) 	
							\\     (name,cs3) 	<- lex  cs
								,  (' ', cs4) 	<- readLitChar cs3
								,  ( x , tail_)	<- reads	   cs4
						 ]
	readsPrec _ = \cs -> [(Con (toString name) x,tail_) 	
							\\     ('(', cs2) 	<- readLitChar cs
								,  (name,cs3) 	<- lex  cs2
								,  (' ', cs4) 	<- readLitChar cs3
								,  ( x , cs5)	<- reads	   cs4
								,  (')', tail_) <- readLitChar cs5
						 ]

instance Read  (Pair a b) | Read a & Read b
where
	readsPrec _ = \cs -> 	[ (Pair x y, tail_) \\ 	( x ,cs2) <- reads cs
												,	(' ',cs3) <- readLitChar cs2
												,	( y ,tail_) <- reads cs3
							]

instance Read (Either a b) | Read a & Read b
where
	readsPrec _ = \cs -> 	[(Left x, res1) \\ (x, res1) <- reads cs] 
						++  [(Right y, res2) \\ (y, res2) <- reads cs]
						

// ----------------------

lexConstr :: String a -> ReadS a
lexConstr const x = \cs -> [(x,res) \\ (le, res) <- lex cs | le == constr ]
	where
		constr :: [Char]
		constr = fromString const

// ---------------------

shown	:: Int a 	  -> [Char] |Show a
shown n x = (showsPrec n x) []

readsn	:: Int -> ReadS a  | Read a 
readsn n = \cs ->  readsPrec n cs

// ************************* Basic type's generic encodings *****************************

// ------ LIST -------------

:: GenList a :== Either (Con Unit) (Con (Pair a [a]))
		
instance toGen [a] (Either (Con Unit) (Con (Pair a [a])))
where
	toGen []		= Left (Con "Nil" Unit)
	toGen [x: xs] 	= Right (Con "Cons" (Pair x xs))

ListToGen :: [a] -> GenList a
ListToGen l = toGen l

// ------ TUPLE -------------

:: GenTuple a b :== Pair a b
		
instance toGen (a,b) (Pair a b)
where
	toGen (x,y) 	= Pair x y

TupleToGen :: (a,b) -> (GenTuple a b)
TupleToGen l = toGen l

// ------ TUPLE -------------

:: GenTriple a b c :== Pair a (Pair b c)
		
instance toGen (a,b,c) (Pair a (Pair b c))
where
	toGen (x,y,z) 	= Pair x (Pair y z)

TripleToGen :: (a,b,c) -> (GenTriple a b c)
TripleToGen l = toGen l

// ------- Instantiations of basic types (Eq, Ord)  ----------------------------------------
/*
instance == [a] | Eq a
where	(==) xs1 xs2 	= (ListToGen xs1) == (ListToGen xs2)

instance < [a] | Ord a
where	(<) xs1 xs2 	= (ListToGen xs1) < (ListToGen xs2)
*/
instance == (a,b) | Eq a & Eq b
where	(==) xs1 xs2 = (TupleToGen xs1) == (TupleToGen xs2)

instance < (a,b) | Ord a & Ord b
where	(<) xs1 xs2  = (TupleToGen xs1) < (TupleToGen xs2)

instance == (a,b,c) | Eq a & Eq b & Eq c
where	(==) xs1 xs2 = (TripleToGen xs1) == (TripleToGen xs2)

instance < (a,b,c) | Ord a & Ord b & Ord c
where	(<) xs1 xs2  = (TripleToGen xs1) < (TripleToGen xs2)

// ----- ORDERING -----------------

instance Show Ordering
where
	showsPrec _ LT	=	showString ['LT']
	showsPrec _ EQ	=	showString ['EQ']
	showsPrec _ GT	=	showString ['GT']

instance Read Ordering
where
	readsPrec _ = \cs -> 		[(o1,res1) \\ (o1,res1) <- (lexConstr "LT" LT) cs] 
							++  [(o1,res1) \\ (o1,res1) <- (lexConstr "EQ" EQ) cs]
							++  [(o1,res1) \\ (o1,res1) <- (lexConstr "GT" GT) cs]

// ----- BOOL ---------------------

instance Show Bool
where
	showsPrec _ True	=	showString ['True']
	showsPrec _ False	=	showString ['False']

instance Read Bool
where
	readsPrec _ = \cs -> 		[(o1,res1) \\ (o1,res1) <- (lexConstr "True" True) cs] 
							++  [(o1,res1) \\ (o1,res1) <- (lexConstr "False" False) cs]

// ----- MAYBE a  -----------------

:: GenMaybe a :== Either (Con Unit) (Con a)

instance toGen (Maybe a) (Either (Con Unit) (Con a))
where
	toGen Nothing	= Left  (Con "Nothing" Unit)
	toGen (Just a) 	= Right (Con "Just"    a)

MaybeToGen :: (Maybe a) -> GenMaybe a
MaybeToGen l = toGen l

MaybeFromGenMaybe :: (GenMaybe a) -> Maybe a
MaybeFromGenMaybe (Left  (Con "Nothing" Unit))	=	Nothing
MaybeFromGenMaybe (Right (Con "Just"    a))		=	Just a

instance Show (Maybe a) | Show a
where
	showsPrec _ Nothing	=	showString ['Nothing']
	showsPrec _ (Just x)=	showString (['Just '] ++ show x)

instance Read (Maybe a) | Read a
where

//	readsPrec _ = \cs -> [(MaybeFromGenMaybe mbg, res) \\ (mbg, res) <- reads cs ]
	readsPrec _ = \cs ->   [(o1,res1) \\ (o1,res1) <- (lexConstr "Nothing" Nothing) cs] 
						++ [(Just x,res2) \\ (jt,res1) <- lex cs, (x, res2) <- reads res1 | jt == ['Just'] ] 
// ----- [a] -----------------

instance Show [a] | Show a
where
	showsPrec _ xs = showList xs

instance Read [a] | Read a
where
	readsPrec _ = readList

// ----- (a,b) -----------------

instance Show (a,b) | Show a & Show b
where
	showsPrec _ (x,y) = showString( ['('] ++ show x ++ [','] ++ show y ++ [')'])

instance Read  (a,b) | Read a & Read b
where
	readsPrec _ = \cs -> 	[ ((x,y), tail_) \\		('(',cs1) <- readLitChar cs
												,	( x ,cs2) <- reads cs1
												,	(',',cs3) <- readLitChar cs2
												,	( y ,cs4) <- reads cs3
												,	(')',tail_) <- readLitChar cs4
							]



// =========================================================================================
