implementation module PreludeIO

from Prelude        import  class Monad(..), class Functor(..), :: Trivial(..), :: Maybe(..), class ==(..), class +++(..)
from PreludeText    import  class Show(..), :: ShowS(..), show, show_
from StdString      import  instance +++ {#Char}
from StdInt         import  instance == Int

import StdFile,StdArray

from StdFunc        import  o
from Prelude        import  error

from PreludeList    import  ++
from PreludeLayer   import  toStr,fromStr,removeNL

// -----------------------------------------------------------------
:: StateMonad  state a  =   MkState (state -> *(a, state))

applyStateMonad     :: (StateMonad .state a) !.state -> (a, !.state)
applyStateMonad stMon =: (MkState stFun) st = stFun st 

appStateMonad       :: (*IOState -> *IOState) -> IO Trivial
appStateMonad stCh  = IO (MkState (\st -> (Trivial, stCh st)))

accStateMonad       :: (*IOState -> *(a, *IOState)) -> IO a
accStateMonad stFun = IO (MkState stFun)

// instance for monads containing state
instance Monad (StateMonad .state)
where
    return val          = (MkState (\st -> (val,st)))

    (>>=) mon1 mon2     = MkState mon3
        where mon3 st = (val2, st2)
                where (val1, st1) = applyStateMonad mon1 st
                      tmon        = mon2 val1
                      (val2, st2) = applyStateMonad tmon st1
    fail s = error s

instance Monad IO
where
    return val  =   IO (MkState (\st -> (val,st)))
    (>>=) (IO mon1_) mon2       = IO (MkState mon3)
        where
            mon3 st = (val2, st2)
            where
                (val1, st1) = applyStateMonad mon1_ st
                tmon        = (conv mon2) val1
                (val2, st2) = applyStateMonad tmon st1
                
    fail s = error s

instance Functor IO
where
    fmap f x = x >>= (return o f)

conv :: (a -> IO b) -> (a -> StateMonad *IOState b)
conv f  =   \a -> fromIO (f a)

fromIO :: (IO a) -> (StateMonad *IOState a)
fromIO (IO a)   = a


// -----------------------------------------------------------------

:: *IOState :== (*World, *File, *(Maybe *File), Maybe Error)
            //  (world, console, opened file,   errors)

:: IO a = IO (StateMonad *IOState a)

:: Error    :== [Char] 
:: FilePath :== [Char]
:: FileMode :== Int

// -----------------------------------------------------------------
getErrorStr :: (Maybe Error) -> String
getErrorStr Nothing     = "\n\n\nNo errors\n" 
getErrorStr (Just str)  = "\n\n\n" +++ toStr str +++ "\n"

// adds new error message to the exsisting list of errors
mkError :: (Maybe Error)  [Char] -> Maybe Error
mkError Nothing     cs  = (Just cs)
mkError (Just cs1) cs2
    # newcs = cs1 ++ ['\n'] ++ cs2
    = (Just newcs)

// opens console, applies the given monad to an initial state, lists the errors
// occured while processing the monad, and prompts for exit
doIO        :: (IO a) *World -> (a, *World)
doIO (IO mon) world
    # (MkState stfun)                   = mon
      (console, world)                  = stdio world
      (val, (world, console, f, err))   = stfun (world, console, Nothing, Nothing)
      console                           = fwrites (getErrorStr err) console
      (_, world)                        = fclose console world
    = (val, world)

// ----- for console --------------------------------------------------------------

// --- Output ---

put_Char :: Char *IOState -> *IOState
put_Char c state =: (w, con, f, err)
    # con   = fwritec c con
    = (w, con, f, err)

putChar :: Char -> IO Trivial
putChar c = appStateMonad (put_Char c)

put_Str :: [Char] *IOState -> *IOState
put_Str c state =: (w, con, f, err)
    # con   = fwrites (toStr c) con
    = (w, con, f, err)

putStr  :: [Char] -> IO Trivial
putStr cs = appStateMonad (put_Str cs)

putStrLn :: [Char] -> IO Trivial
putStrLn cs = appStateMonad (put_Str ( cs ++ ['\n']))

print :: a -> IO Trivial | Show a
print a = appStateMonad (put_Str ( (show a) ++ ['\n']))

// --- Input ---

get_Char :: *IOState -> (Char, *IOState)
get_Char state =: (w, con, f, err)
    # (ok, c, con)  = freadc con
    | ok    = (c, (w, con, f, err))
            = (c, (w, con, f, mkError err ['Can not get char.']))

getChar :: IO Char
getChar = accStateMonad get_Char

get_Line :: *IOState -> ([Char], *IOState)
get_Line state =: (w, c, f, err)
    # (str, c)      = freadline c
    | size str == 0     
        =   (fromStr str, (w, c, f,mkError err ['Can not get line.']))
        #   cs = fromStr (removeNL str)
        = (cs, (w, c, f, err))

getLine :: IO [Char]
getLine = accStateMonad get_Line

get_Contents :: *IOState -> ([Char], *IOState)
get_Contents state =: (w, con, f, err) 
    #   (con,content) = get_Contents` con []
    = (content, (w, con, f, err))

get_Contents` :: *File [Char] -> (*File ,[Char])
get_Contents` con buf
    #   (ok, c, con)    = freadc con
    | ok    =   get_Contents` con (buf ++ [c])
            =   (con,buf)

getContents :: IO [Char]
getContents = accStateMonad get_Contents

interact_ :: ([Char] -> [Char]) *IOState -> *IOState
interact_ strfun state =: (w, con, f, err)
    #   (con,content)   = get_Contents` con []
        con             = fwrites (toStr (strfun content)) con
    = (w, con, f, err)

interact :: ([Char] -> [Char]) -> IO Trivial
interact strfun = appStateMonad (interact_ strfun)

// ----- for file ----------------------------------------------------------

// opens a file; if there already is an open file, gives an error message
open_file :: FilePath Int *IOState -> *IOState
open_file fn m state =: (w, c, (Just f), err)
    # err = mkError err ['There is already an opened file.']            //!!! baj lehet belole
    = (w, c, Nothing, err)
open_file fn m state =: (w, c, Nothing, err)
    # (ok, f, w) = fopen (toStr fn) m w
    | ok
    = (w, c, (Just f), err)
        # err = mkError err (['Unable to open file '] ++ fn ++ ['.'])
        = (w, c, Nothing, err)

// closes the opened file; gives an error message if there's no open file
close_file  :: *IOState -> (Trivial, *IOState)
close_file state =: (w, c, Nothing, err)
    # err = mkError err ['No file to close.']
    = (Trivial, (w, c, Nothing, err))
close_file state =: (w, c, (Just f), err)
    # (ok, w)   = fclose f w
    | ok
    = (Trivial, (w, c, Nothing, err))
        # err = mkError err ['Unable to close file.']
        = (Trivial, (w, c, Nothing, err))

// --- Input ---

read_File :: FilePath *IOState -> ([Char], *IOState)
read_File fp state 
    #   state                   =   open_file fp FReadText state
        (w, con, f, err)    =   state
    =   fun w con f err
        /*
        case f of
            (Just fl)    
                #   (fl,content) = get_Contents` fl []
                ->  (content,(w,con,(Just fl),err,cl))
            Nothing     ->  ([], state)
        */
    where
        fun w con (Just fl) err
            #   (fl,content)    = get_Contents` fl []
                (_,state)       = close_file (w,con,(Just fl),err)
            =   (content,state)
        fun w con Nothing err   = ([], (w,con,Nothing,err))

readFile :: FilePath -> IO [Char]
readFile fp = accStateMonad (read_File fp)

// --- Output ---

write_File :: FilePath [Char] *IOState -> *IOState
write_File fp cs state 
    #   state                   =   open_file fp FWriteText state
        (w, con, f, err)    =   state
    =   fun w con f err
    where
        fun w con (Just fl) err
            #   fl              = fwrites (toStr cs) fl
                (_,state)       = close_file (w,con,(Just fl),err)
            =   state
        fun w con Nothing err   = (w,con,Nothing,err)


writeFile :: FilePath [Char] -> IO Trivial
writeFile fp cs = appStateMonad (write_File fp cs)

append_File :: FilePath [Char] *IOState -> *IOState
append_File fp cs state 
    #   state                   =   open_file fp FAppendText state
        (w, con, f, err)    =   state
    =   fun w con f err 
    where
        fun w con (Just fl) err
            #   fl              = fwrites (toStr cs) fl
                (_,state)       = close_file (w,con,(Just fl),err)
            =   state
        fun w con Nothing err   = (w,con,Nothing,err)

appendFile :: FilePath [Char] -> IO Trivial
appendFile fp cs = appStateMonad (append_File fp cs)
