implementation module PreludeText

from PreludeList import ++, dropWhile,elem,span,map,!!, length
from StdFunc     import o
from StdChar     import isAlphanum

from StdReal        import instance fromReal {#Char}
from StdList        import instance fromString []
from Ratio          import %
from Char           import isSpace,isAlpha,isDigit,isOctDigit,isHexDigit,isUpper
                          ,showLitChar,readLitChar,lexLitChar
from Numeric        import showSigned, showInt, readSigned, readDec, lexDigits

import Prelude
// ---------------------------------------------------------------------------------------------------

:: ReadS a  :== [Char] -> [(a,[Char])]
:: ShowS    :== [Char] -> [Char]

// -------- READ class ---------    
class   Read a  
where
    readsPrec   :: Int -> ReadS a

    readList    :: ReadS [a] | Read a
    readList    :== readList_

readList_ :: ReadS [a] | Read a
readList_ = readParen False (\r -> [pr  \\  (['['],s)  <- lex r,    pr  <- readl s])
        where 
            readl s =   [([],t)     \\  ([']'],t)  <- lex s] 
                        ++
                        [([x:xs],u) \\  (x,t)    <- reads s, (xs,u) <- readl` t]
            readl` s=   [([],t)     \\  ([']'],t) <- lex s] 
                        ++
                        [([x:xs],v) \\  ([','],t)  <- lex s, (x,u)    <- reads t, (xs,v)   <- readl` u]

class   Show a
where
    showsPrec   :: Int a    -> ShowS
    //showsPrec _ x s   :== show x ++ s
    
    show        :: a -> [Char]  | Show a
    show x      :== (showsPrec 0 x) []
    
    showList    :: [a] -> ShowS | Show a
    showList xs :== showList_ xs

showList_ :: [a] -> ShowS   | Show a
showList_ []        = showString ['[]']
showList_ [x:xs]    = showChar '[' o shows x o showl xs
    where 
        showl []        = showChar ']'
        showl [x:xs]    = showChar ',' o shows x o showl xs

show_     :: a -> ShowS | Show a
show_ x   = showsPrec 0 x

// -------------

reads   :: ReadS a      | Read a
reads   = readsPrec 0

shows   :: a -> ShowS   | Show a
shows a = showsPrec 0 a

read    :: [Char] -> a  | Read a
read s  =  
    case [x \\ (x,t) <- reads s, ([],[]) <- lex t] of
            [x] -> x
            []  -> error ['read: no parse']
            _   -> error ['read: ambiguous parse']

showChar    :: Char -> ShowS
showChar c = \cs ->  [c:cs]

showString  :: [Char] -> ShowS
showString str  = \cs  -> str ++ cs

showParen   :: Bool ShowS -> ShowS
showParen b p   =  if b (showChar '(' o p o showChar ')') p

readParen   :: Bool (ReadS a) -> ReadS a
readParen b g   =  
    if b mandatory optional
    where 
        optional r  = g r ++ mandatory r
        mandatory r = [(x,u)    \\  (['('],s) <- lex r, (x,t)   <- optional s,  ([')'],u) <- lex t  ]

lex             :: ReadS [Char]
lex = lex`
    where
        lex` []         =   [([],[])]
        lex` [c:s]
           | isSpace c  =   lex (dropWhile isSpace s)
        lex` ['\'':s]   =   [ (['\'':ch++['\'']], t)    \\ (ch,['\'':t])  <- lexLitChar s   //  'c' : character
                                                        |   ch /= ['\'']    ]
        lex` ['"':s]    =   [(['"':str], t) \\ (str,t) <- lexString s]                      // "string"
            where
                lexString ['"':s]       = [(['\"'],s)]                                          // empty string
                lexString s             =   [   (ch++str, u)    \\  (ch,t)  <- lexStrItem s, (str,u) <- lexString t     ]
                lexStrItem ['\\':['&':s]]   =  [(['\\&'],s)]            // '\&' is a null charachter (used for : "\65\&9" == "a9")
                lexStrItem ['\\':[c:s]] // Strings could be written in several lines, an '\' should be at the end of the line and another one at the beginning of the next line
                    | isSpace c         =   [   (['\\&'],t) \\  ['\\':t] <- [dropWhile isSpace s]   ]
                lexStrItem s            =   lexLitChar s
        lex` [c:s] 
            | isSingle c    =   [([c],s)]
            | isSym c       =   [([c:sym],t)    \\  (sym,t) <- [span isSym s]   ]
            | isAlpha c     =   [([c:nam],t)    \\  (nam,t) <- [span isIdChar s]    ]
            | isDigit c     =   [([c:ds++fe],t) \\  (ds,s)  <- [span isDigit s],    (fe,t)  <- lexFracExp s ]
                            = []    // bad character
            where
                isSingle c  = elem c [',;()[]{}_`']
                isSym c     = elem c ['!@#$%&*+./<=>?\\^|:-~']
                isIdChar c  = isAlphanum c || elem c ['_','\''] //2.0['_\'']
                lexFracExp ['.':[c:cs]] 
                    | isDigit c =   [   (['.':ds++e],u) \\  (ds,t)  <- lexDigits [c:cs] , (e,u) <- lexExp t ]
                lexFracExp s        =   [([],s)]

                lexExp [e:s] 
                    | elem e ['eE']
                        =   [([e:[c:ds]],u) \\  [c:t]  <- [s], (ds,u) <- lexDigits t
                                            |   elem c ['+-']   ] // e.g. e-56
                            ++
                            [([e:ds],t)     \\ (ds,t) <- lexDigits s ]  // e.g. e56
                lexExp s = [([],s)]

// ****** INT *******

instance Show Int 
where showsPrec p n =   showSigned showInt p n

instance Read Int
where readsPrec p   =   readSigned readDec

// ****** REAL *******

instance Show Real
where 
    showsPrec :: Int Real -> ShowS
    showsPrec p r   
        | (-1.0) < r && r < 0.0 =   showString ( ['-0'] ++ fromString (toString (~r)))
        |    0.0 < r && r < 1.0 =   showString (  ['0'] ++ fromString (toString r))
                                =   showString (fromString (toString r))

instance Read Real 
 where 
    readsPrec p = readFloat 

// ****** TRIVIAL *******

instance Show Trivial  where    showsPrec p Trivial = showString ['()']

instance Read Trivial
where
    readsPrec p    = readParen False (\r -> [(Trivial,t) \\ (['('],s) <- lex r, ([')'],t) <- lex s ] )

// ****** CHAR *******

instance  Show Char
    where
        showsPrec p '\'' = showString ['\'','\\','\'','\'']
        showsPrec p c    = showChar '\'' o showLitChar c o showChar '\''

//instead of showList member of Show (which is predefined)
showCharList :: [Char] -> ShowS
showCharList cs = showChar '"' o showl cs
    where 
        showl []        = showChar '"'
        showl ['"':cs]  = showString ['\\\"'] o showl cs
        showl [c:cs]    = showLitChar c o showl cs


instance  Read Char
where
    readsPrec p = readParen False   (\r -> [(c,t) \\ (['\'':s],t)<- lex r, (c,['\''])  <- readLitChar s] )

//instead of readList member of Read (which is predefined)
readCharList :: [Char] -> [([Char],[Char])]
readCharList cs
    = readParen False (\r -> [(l,t) \\ (['"':s], t) <- lex r,   (l,_) <- readl s ]) cs
    where
        readl ['"':s]           = [([],s)]
        readl ['\\':['&':s]]    = readl s
        readl s                 = [([c:cs],u) \\ (c ,t) <- readLitChar s, (cs,u) <- readl t ]

/* See in PreludeGeneric:
instance Show (a,b)| Show a & Show b
nstance  Read (a,b)| Read a & Read b
*/

// ------------------------ from NUMERIC library --------------------------------------------------------

readFloat 
    = \r -> [   ( fromRational (n % (toEnum 1) ) * (toEnum  10) ^^ (k-d)
                , t) 
            \\  (n,d,s) <- readFix r,   (k,t)   <- readExp s
            ]
    where
        readFix r = [(read (ds++ds`), length ds`, t)    \\  (ds,d) <- lexDigits r,  (ds`,t) <- lexFrac d ]

        lexFrac ['.':ds] = lexDigits ds
        lexFrac s        = [([],s)]        

        readExp [e:s] | elem e ['eE']   = readExp` s
        readExp s                       = [(0,s)]

        readExp` ['-':s]    = [(negate k,t) \\ (k,t) <- readDec s]
        readExp` ['+':s]    = readDec s
        readExp` s          = readDec s

